 //===================================================
//
// File Model.java
// jlouie 1.98
// Encapsulates application logic and data
// Supplied "as is"
// No warranty is expressed or implied
// This code is for instructional use only
// This version supports notification of observers
// when the state of the Model object changes.
// This version supports assertions and does not
// validate input if Assert.IS_ASSERTIONS_ON is false.
//
//===================================================

import java.util.*;
import java.awt.*;
import java.io.*;

public class Model extends Observable implements Serializable {
    
    // implement Serializable
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException{
        in.defaultReadObject();
    }
    
    private void writeObject(ObjectOutputStream out)throws IOException {
        out.defaultWriteObject();
    }
    
    // class constants
    public static final int INVALID= -1;
    public static final int PRINCIPAL= 0;
    public static final int INTEREST= 1;
    public static final int MONTHS= 2;
    public static final int PAYMENT= 3;
    
    // array of double, stores numeric input
    private double    arrayDb[]= new double[4];
    private int       target= INVALID;
    private String    message="";
    
    // self test
    public static void main(String[] args) throws Exception {
        Model model= new Model(100000,8.5,360,0);
        System.out.println("Message: "+model.getMessage());
        System.out.println("Result: "+model.getResult());
        System.out.println("Principal: "+model.getPrincipal());
        System.out.println("Interest: "+model.getInterest());
        System.out.println("Months: "+model.getMonths());
        System.out.println("Payment: "+model.getPayment());
        System.in.read();
    }
    
    public Model(){
        for (int i=0; i<4; i++) {arrayDb[i]= 0.0;}
    }
   
    public  Model(double principal, double interest,
                            int months, double payment){
        init(principal, interest, months, payment);                          
    }
    
    public void init(double principal, double interest,
                            int months, double payment) {
        target= INVALID;
        message= "Done";
        arrayDb[PRINCIPAL]= principal;
        arrayDb[INTEREST]= interest;
        arrayDb[MONTHS]= (double)months;
        arrayDb[PAYMENT]= payment;
        
        
        // ASSERT one and only one input is zero
        // ASSERT no value may be negative
        // ASSERT interest less than or equal to 100
        // ASSERT months is integer value
        
        
        boolean isAnyNegative= false;
        int zeros= 0;
        int temp= INVALID;   
        for (int i=0; i<4; i++) {
            if (arrayDb[i] == 0 ) {
                zeros +=1;  
                temp= i;
            }
            if (arrayDb[i]<0) {isAnyNegative= true;}
        }
        
        //if (Assertion.IS_ASSERTIONS_ON) {
            Assertion.assert(! isAnyNegative, "Invalid negative parameter in Model.init().");
            Assertion.assert(zeros == 1,"One and only one parameter must be zero in Model.init().");
            Assertion.assert(arrayDb[INTEREST]<=100, "Invalid interest in Model.init().");
        //}
        // processing does not get here if assertion fails.
        target= temp;
        doAlgo(target);
        doUpdate();
        
        // old bulletproofing code, but inefficient. Calling class must verify pre-conditions.
        /*try {
            if (zeros>1){throw new Exception("Too many zeros.");}
            if (zeros == 0){throw new Exception("One entry must be zero.");}
            if (isAnyNegative) {throw new Exception("Invalid negative value.");}
            if (arrayDb[INTEREST]>100){throw new Exception("Invalid interest.");}
            if ((int)arrayDb[MONTHS] != arrayDb[MONTHS]){throw new Exception("Invalid months.");}
            target= temp;
            doAlgo(target);  // finally, do the math!
            doUpdate();  // notify observers
        }
        catch(Exception e) {
            message= e.getMessage();  
            // clear the array
            for (int i=0; i<4; i++) {arrayDb[i]= 0.0;}
            target= INVALID;
        }*/
        
    }

   //==================================
   // * doAlgo * calcInt *
   //
   // The actual amortization algorithm:
   //
   // m= P*i/(1-(1+i)^-N)
   // i=r/1200
   //
   // solving for i is non_trivial
   // Thanks to Dr. W. Carlini (& Newton) for this formula!
   //
   //==================================
   
   private double calcInt(double P, double N, double m){

      double temp=(m/P), answer=(m/P), diff=100, num=0, den=0, accuracy=.00001;
      int i, maxIterations= 1000;
      
      try{
         for (i=0; ((diff>accuracy) && (i<maxIterations)); i++){
            temp= answer;
            num= (P*temp/m)+Math.pow((1+temp), -N)-1;
            den= (P/m)-N*Math.pow((1+temp),(-N-1));
            //if (den==0) throw new ArithmeticException(); // not needed
            answer= temp- (num/den);
            diff= answer- temp;
            if (diff<0) diff= -diff;
         }
         answer *= 1200;
         if ((answer < 0) || (Double.isNaN(answer)) || (i == maxIterations)){
            throw new ArithmeticException();
         }
      }
      catch (ArithmeticException e) {answer= 0;}
      return answer;
   }
   

    private void doAlgo(int field) {
        String   outString;
        double   db=0, temp= 0;
        double   P= arrayDb[0];          // principal
        double   i= (arrayDb[1]/1200);   // monthly percentage rate
        double   N= arrayDb[2];          // loan period months
        double   m= arrayDb[3];          // monthly payments    

        if (field<4){
            try{
                switch (field){
                    case 0: // Principal
                        db= 1+i;
                        db= 1/Math.pow(db, N);
                        db= ((1-db)/i)*m;
                        break;
                    case 1:  // Annual Interest
                        if ((N*m)<P) throw new ArithmeticException();
                        db= calcInt(P, N, m);
                        break;
                    case 2:  // Loan Period
                        db= (1-(P*i/m));
                        db= Math.log(db);
                        temp= 1+i;
                        db= -db/Math.log(temp);
                        break;
                    case 3:  // Monthly Payments
                        db= 1+i;
                        db= 1/Math.pow(db, N);
                        db= (P*i)/(1-db);
                        break;
                    default:;
                }
            }
            catch (ArithmeticException e) {db=0;}
        }
        if (Double.isNaN(db)) db= 0;
        if (db == 0) {message ="Input Error";}
        else arrayDb[target]=db;
    }
    
    // allow users of this class to initiate updates
    // if data is invalid, Hashtable will contain values "0.0"
    public void doUpdate() {
        Hashtable hash= new Hashtable();
        hash.put("principal",String.valueOf(getPrincipal()));
        hash.put("interest",String.valueOf(getInterest()));
        hash.put("months",String.valueOf(getMonths()));
        hash.put("payment",String.valueOf(getPayment()));
        hash.put("message",message);
        setChanged();
        notifyObservers(hash);       
    }
   
    // default target= -1;
    public boolean isValid() {
        return ((target>=PRINCIPAL) && (target<=PAYMENT)) ? true: false;  
    }
    
    // get calculated value
    public double getResult() {return arrayDb[target];}
    public int    getTarget() {return target;}
    public String getMessage() {return message;}
    public double getPrincipal() {return (isValid()) ? arrayDb[PRINCIPAL]: 0.0;}
    public double getInterest() {return (isValid()) ? arrayDb[INTEREST]: 0.0;}
    public double getMonths() {return (isValid()) ? arrayDb[MONTHS]: 0;}
    public double getPayment() {return (isValid()) ? arrayDb[PAYMENT]: 0.0;}
}
